#include <gtest/gtest.h>

#include "defs.h"

namespace defs {
namespace {

// TODO: Actually add unit tests here.

// TEST(Defs, IsAux) {
//   EXPECT_FALSE(IsAux(/*threadpos=*/26, /*npml=*/5));
//   EXPECT_TRUE(IsAux(/*threadpos=*/27, /*npml=*/5));
//
//   RunShape rs(/*blockshape=*/UV(0, 0),
//               /*gridshape=*/UV(0, 0),
//               /*spacing=*/0,
//               /*domain=*/XY(0, 0),
//               /*srctype=*/RunShape::SourceType::ZSLICE,
//               /*nout=*/0,
//               /*npml=*/5);
//
//   EXPECT_FALSE(IsAux(/*threadpos=*/26, rs));
//   EXPECT_TRUE(IsAux(/*threadpos=*/27, rs));
// }

} // namespace
} // namespace defs
